@@ -4,7 +4,7 @@
 #  by Brian Warner <warner@lothar.com>
 #  Released Version 0.6
 #
-#  $Id: egd.pl,v 1.38 1999/08/05 06:37:59 warner Exp $
+#  $Id: egd.pl,v 1.44 2000/04/03 06:49:39 warner Exp $
 
 # This is a perl program to collect entropy from a running Unix-like system,
 # mostly through various stats calls (vmstat and friends). The output is
@@ -34,18 +34,23 @@ use IO::File;
 use Getopt::Long;
 
 sub Usage {
-    print "egd.pl [options] /tmp/entropy\n";
-    print "Options:\n";
-    print " --debug: display entropy count, sleep time on each loop pass\n";
-    print " --debug-select: print status of fds used in select() loop\n";
-    print " --debug-gather: print stats of gatherers started and reaped\n";
-    print " --debug-client: print info about client requests\n";
-    print " --debug-pool: print size, contents of entropy pool\n";
-    print " --bottomless: for testing, don't decrement entropy count\n";
-    print " --nofork: do not put self in background: remain in foreground\n";
-    print " --quit-after=30: terminate after 30 minutes\n";
-    print "\n";
-    print " --kill: kill off the local daemon on other side of the socket\n";
+    print STDERR <<'EOF';
+egd.pl [options] /tmp/entropy
+Options:
+ --debug: display entropy count, sleep time on each loop pass
+ --debug-select: print STDERR status of fds used in select() loop
+ --debug-gather: print STDERR stats of gatherers started and reaped
+ --debug-client: print STDERR info about client requests
+ --debug-pool: print STDERR size, contents of entropy pool
+ --bottomless: for testing, don't decrement entropy count
+ --nofork: do not put self in background: remain in foreground
+ --quit-after=30: terminate after 30 minutes
+
+ --kill: kill off the local daemon on other side of the socket
+
+ socket is either a TCP socket (8000 or localhost:8000), or a pathname to a
+ UNIX-domain socket (/tmp/entropy, ~/.gnupg/entropy, foo.entropy).
+EOF
     exit(0);
 }
 
@@ -99,10 +104,13 @@ sub get_pid_from_remote {
 
 if ($kill_mode) {
     foreach my $socket (@ARGV) {
-	next unless $socket =~ m!^/!;
+	if ($socket =~ /^\d+$/ or $socket =~ /:/) {
+	    print STDERR " cannot --kill across TCP sockets for '$socket'\n";
+	    next;
+	}
 	my $pid = get_pid_from_remote($socket);
 	next unless $pid;
-	print "killing process $pid\n";
+	print STDERR "killing process $pid\n";
 	kill 1, ($pid);
 	sleep(1);
 	kill 9, ($pid);
@@ -111,9 +119,9 @@ if ($kill_mode) {
 }
 
 unless (@ARGV) {
-    print "No sockets to read entropy: not very useful.\n";
-    print "Try running as:  egd.pl ~/.gnupg/entropy\n";
-    print "Terminating.\n";
+    print STDERR "No sockets to read entropy: not very useful.\n";
+    print STDERR "Try running as:  egd.pl ~/.gnupg/entropy\n";
+    print STDERR "Terminating.\n";
     exit(0);
 }
 
@@ -125,11 +133,14 @@ unless (@ARGV) {
 # check to see if a daemon is already running on a socket. If so, remove that
 # socket from the list
 foreach my $socket (@ARGV) {
-    if ($socket =~ m!^/!) {
+    if ($socket =~ /^\d+$/ or $socket =~ /:/) {
+	# TCP socket
+	push(@lsockets, [$socket,0]);
+    } else {
+	# unix socket
 	my $pid = get_pid_from_remote($socket);
-	next if $pid; # already running
+	push(@lsockets, [$socket,0]) unless $pid; # already running
     }
-    push(@lsockets, [$socket,0]) if $socket =~ /^\// or $socket =~ /^\d+$/;
 }
 
 if (!@lsockets) {
@@ -144,8 +155,6 @@ if (!@lsockets) {
 # these are added once the daemon starts up:
 #  $s[2]: the IO::Socket object
 
-# get socket names with --socket and --writesocket
-
 $pool = "\x00" x $POOLSIZE;
 
 $next_source = 0;
@@ -234,6 +243,9 @@ bpb=0.1,filter=1
 bpb=0.3,filter=0
  netstat -n
 
+bpb=0.2,filter=0
+ lsof
+
 #if defined( __sgi ) || defined( __hpux )
 bpb=0.1,filter=0 /bin/ps -el
 #endif				/* __sgi || __hpux */
@@ -363,13 +375,13 @@ EOF
 	}
     }
     die "bad source list: no sources" unless @sources;
-    print scalar(@sources)," sources found\n";
+    print STDERR scalar(@sources)," sources found\n";
 }
 
 if (0) {
-    print "SOURCES:\n";
+    print STDERR "SOURCES:\n";
     foreach my $s (@sources) {
-	print join(',', 
+	print STDERR join(',', 
 		   map {"$_->$s->{$_}"} sort(keys(%$s)) ),"\n";
     }
 }
@@ -382,10 +394,11 @@ unless ($nofork) {
     # reopen stdin,out,err so they're not connected to the old tty
     # background ourselves
 
-    print "forking into background...\n";
+    print STDERR "forking into background...\n";
 
     # for now, just do the last one. I'd like some error/debug messages to
-    # have somewhere to go for a while.
+    # have somewhere to go for a while, and the sockets might be relative
+    # to the current directory.
     fork && exit(0);
 }
 
@@ -395,7 +408,7 @@ sub add_entropy {
     # stir in data
     my $c = new SHA;
     while(length($data)) {
-	my $chunk = substr($pool, 0, $DIGEST_SIZE);
+	my $chunk = substr($data, 0, $DIGEST_SIZE);
 	$pool ^= $chunk;
 	stir();
 	substr($data, 0, $DIGEST_SIZE) = "";
@@ -431,14 +444,14 @@ sub stir {
 	my $d = $c->digest();
 	substr($pool, $i, $DIGEST_SIZE) = substr($pool, $i, $DIGEST_SIZE) ^ $d;
     }
-    print "pool: ",unpack("H*",$pool),"\n" if $debug_pool;
+    print STDERR "pool: ",unpack("H*",$pool),"\n" if $debug_pool;
 }
 
 sub shutdown_thing {
     my($c) = @_;
-    print "shutdown_thing($c)\n" if $debug_select;
+    print STDERR "shutdown_thing($c)\n" if $debug_select;
     if ($c->{'type'} eq 'gatherer') {
-	print "reaping pid ",$c->{'pid'}," source ",$c->{'source'},"\n"
+	print STDERR "reaping pid ",$c->{'pid'}," source ",$c->{'source'},"\n"
 	   if $debug_gather;
 	kill 9, $c->{'pid'};
 	waitpid($c->{'pid'},0); # portability check: &WNOHANG?
@@ -447,28 +460,30 @@ sub shutdown_thing {
 	if ($?) {
 	    # non-zero exit: error. Mark that source as non-usable
 	    $sources[$c->{'source'}]->{'unusable'} = 1;
-	    print " marked as nonusable, rc == $?\n" if $debug_gather;
+	    print STDERR " marked as nonusable, rc == $?\n" if $debug_gather;
 	} elsif (!defined($c->{'incoming'}) or !length($c->{'incoming'})) {
 	    # no stdout, probably all stderr, mark unusable
 	    $sources[$c->{'source'}]->{'unusable'} = 1;
-	    print " marked as nonusable, empty stdout\n" if $debug_gather;
+	    print STDERR " marked as nonusable, empty stdout\n" 
+	      if $debug_gather;
 	} else {
 	    # add in the entropy from that source
 	    #  possibly filter it first
 	    if ($sources[$c->{'source'}]->{'filter'}) {
-		print " filtered from ",length($c->{'incoming'})," bytes\n"
-		  if $debug_gather;
+		print STDERR " filtered from ",length($c->{'incoming'}),
+		  " bytes\n" if $debug_gather;
 		$c->{'incoming'} =~ s/\D//g;
 	    }
 	    my $bytes = length($c->{'incoming'});
 	    my $source = $sources[$c->{'source'}];
 	    my $entropy = $bytes * $source->{'bpb'};
-	    print " returned $bytes bytes, $entropy bits\n" if $debug_gather;
+	    print STDERR " returned $bytes bytes, $entropy bits\n" 
+	      if $debug_gather;
 	    $entropy = $source->{'maxbits'}
 	      if $entropy > $source->{'maxbits'};
-	    print " accepted $entropy bits\n" if $debug_gather;
+	    print STDERR " accepted $entropy bits\n" if $debug_gather;
 	    add_entropy($c->{'incoming'}, $entropy);
-	    print scalar(@want_entropy)," clients want entropy\n"
+	    print STDERR scalar(@want_entropy)," clients want entropy\n"
 	      if $debug_client;
 	    foreach my $w (@want_entropy) {
 		my $bytes = $w->{'wanted'};
@@ -525,38 +540,36 @@ $refresh_time = time() + $REFRESH_TIME;
 $read_select = new IO::Select;
 $error_select = new IO::Select;
 
-print "server starting\n";
+print STDERR "server starting\n";
 foreach my $lsocket (@lsockets) {
     my $where = $lsocket->[0];
     my $s;
-    if ($where =~ m!^/!) {
-	print "listening on unix socket at $where\n" if $debug_client;
+    if ($where =~ /^\d+$/) {
+	print STDERR "listening on tcp socket at $where\n" if $debug_client;
+	$s = new IO::Socket::INET (
+				   'LocalAddr' => $where,
+				   'Proto' => 'tcp',
+				   'Listen' => 5,
+				   'Timeout' => 0,
+				   'Reuse' => 1,
+				  );
+    } elsif ($where =~ /:/) {
+	print STDERR "listening on tcp socket port $where\n" if $debug_client;
+	$s = new IO::Socket::INET (
+				   'LocalPort' => $where,
+				   'Proto' => 'tcp',
+				   'Listen' => 5,
+				   'Timeout' => 0,
+				   'Reuse' => 1,
+				  );
+    } else {
+	print STDERR "listening on unix socket at $where\n" if $debug_client;
 	unlink($where);
 	$s = new IO::Socket::UNIX (
 				   'Type' => SOCK_STREAM,
 				   'Local' => $where,
 				   'Listen' => 1,
 				  );
-    } else {
-	if ($where =~ /:/) {
-	    print "listening on tcp socket at $where\n" if $debug_client;
-	    $s = new IO::Socket::INET (
-				       'LocalAddr' => $where,
-				       'Proto' => 'tcp',
-				       'Listen' => 5,
-				       'Timeout' => 0,
-				       'Reuse' => 1,
-				      );
-	} else {
-	    print "listening on tcp socket port $where\n" if $debug_client;
-	    $s = new IO::Socket::INET (
-				       'LocalPort' => $where,
-				       'Proto' => 'tcp',
-				       'Listen' => 5,
-				       'Timeout' => 0,
-				       'Reuse' => 1,
-				      );
-	}
     }
     die "can't create socket $where : $!" unless $s;
     $s->listen or die("couldn't listen on socket $where : $!");
@@ -589,7 +602,7 @@ while (1) {
 	# by expire time, and the next-to-expire gatherer is used to
 	# determine the timeout for the select call
 	if (@gatherers > $MAX_GATHERERS) {
-	    print "throttling gatherers\n" if $debug_gather;
+	    print STDERR "throttling gatherers\n" if $debug_gather;
 	    goto done_spawn;
 	}
 	my $start = $next_source;
@@ -608,13 +621,13 @@ while (1) {
 		# but should be caught elsewhere) or we've used the sources too
 		# much. In that case, arrange to sleep until one of the sources
 		# becomes useable again.
-		print "ran out of sources\n" if $debug_gather;
+		print STDERR "ran out of sources\n" if $debug_gather;
 		if (@gatherers) {
 		    goto done_spawn;
 		}
 		# find out how long we have to wait until we can run a new
 		# source
-		print " waiting for a source to become usable again\n" 
+		print STDERR " waiting for a source to become usable again\n" 
 		  if $debug_gather;
 		$next_source = ($next_source +1) % @sources;
 		while ($next_source != $start) {
@@ -628,8 +641,8 @@ while (1) {
 		    }
 		    $next_source = ($next_source +1) % @sources;
 		}
-		print " next available is in ",($next_available - time()),"\n"
-		  if $debug_gather;
+		print STDERR " next available is in ",
+		  ($next_available - time()),"\n" if $debug_gather;
 		goto done_spawn;
 	    }
 	}
@@ -661,11 +674,11 @@ while (1) {
 	      || exit(1);	#die "can't exec program: $!"; # not so noisy
 	}
 	push(@gatherers, $g);
-	print "spawned src $s (",$sources[$s]->{'cmd'}
+	print STDERR "spawned src $s (",$sources[$s]->{'cmd'}
 	  if $debug_gather;
-	print ") at ",time(),", expire in ",$sources[$s]->{'timeout'}
+	print STDERR ") at ",time(),", expire in ",$sources[$s]->{'timeout'}
 	  if $debug_gather;
-	print " at ",$g->{'expire'},", fd $fd\n" if $debug_gather;
+	print STDERR " at ",$g->{'expire'},", fd $fd\n" if $debug_gather;
       done_spawn:
     }
 
@@ -678,30 +691,32 @@ while (1) {
     @gatherers = sort { $a->{'expire'} <=> $b->{'expire'} } @gatherers;
     if (@gatherers) {
 	my $expire = $gatherers[0]->{'expire'} - time();
-	print "gatherer ",$gatherers[0]->{'source'}," is next to expire "
-	  if $debug_gather;
-	print "in $expire secs\n" if $debug_gather;
+	print STDERR "gatherer ",$gatherers[0]->{'source'},
+	  " is next to expire " if $debug_gather;
+	print STDERR "in $expire secs\n" if $debug_gather;
 	$timeout = $expire if $timeout > $expire;
     }
     if (defined($next_available)
 	and ($next_available - time() < $timeout)) {
 	$timeout = $next_available - time();
-	print " next_available being used\n" if $debug_gather;
+	print STDERR " next_available being used\n" if $debug_gather;
     }
     $timeout += 1;
 
     if ($debug_select) {
-	print "select: r(",join(',',
+	print STDERR "select: r(",join(',',
 				$read_select->handles()),"), ";
-	print "w(",join(',',$write_select->handles()),"), ";
-	print "e(",join(',',$error_select->handles()),"),";
-	print "t=$timeout, entropy=$pool_entropy\n";
+	print STDERR "w(",join(',',$write_select->handles()),"), ";
+	print STDERR "e(",join(',',$error_select->handles()),"),";
+	print STDERR "t=$timeout, entropy=$pool_entropy\n";
     }
 
     if ($debug) {
-	print "e: $pool_entropy ($MIN_ENTROPY,$MAX_ENTROPY), g:$gathering";
-	print " na:",$next_available - time() if defined($next_available);
-	print "\n";
+	print STDERR 
+	  "e: $pool_entropy ($MIN_ENTROPY,$MAX_ENTROPY), g:$gathering";
+	print STDERR " na:",$next_available - time()
+	  if defined($next_available);
+	print STDERR "\n";
     }
 
     # wait for input: blocks here until work can be done
@@ -709,11 +724,11 @@ while (1) {
       IO::Select::select($read_select, $write_select, $error_select,
 			 $timeout);
     if ($debug_select) {
-	print " return: ";
-	print "r(",$readable ? join(',',@$readable) : "nil","), ";
-	print "w(",$writable ? join(',',@$writable) : "nil","), ";
-	print "e(",$errorful ? join(',',@$errorful) : "nil","),";
-	print "t=$timeout\n";
+	print STDERR " return: ";
+	print STDERR "r(",$readable ? join(',',@$readable) : "nil","), ";
+	print STDERR "w(",$writable ? join(',',@$writable) : "nil","), ";
+	print STDERR "e(",$errorful ? join(',',@$errorful) : "nil","),";
+	print STDERR "t=$timeout\n";
     }
 
     # if the quit time has arrived, quit
@@ -723,7 +738,7 @@ while (1) {
 
     # if we timed out, reduce the entropy count a little bit
     if (time() > $refresh_time) {
-	print "timed out, decaying entropy\n" if $debug;
+	print STDERR "timed out, decaying entropy\n" if $debug;
 	$pool_entropy -= $DECAY if $pool_entropy > $DECAY;
 	stir();
 	$refresh_time = time() + $REFRESH_TIME
@@ -735,7 +750,7 @@ while (1) {
 	    die "problem: lsocket had error\n";
 	}
 	# shut it down, kill it off, reap it
-	print "shutting down errorful socket $s\n" if $debug_select;
+	print STDERR "shutting down errorful socket $s\n" if $debug_select;
 	my $c = $s->[1];
 	shutdown_thing($c);
     }
@@ -746,7 +761,8 @@ while (1) {
 	my $out = \$c->{'outgoing'};
 	my $nwritten = syswrite($c->{'fd'}, $$out, length($$out));
 	unless ($nwritten) {
-	    print "error while writing, shutting down\n" if $debug_select;
+	    print STDERR "error while writing, shutting down\n"
+	      if $debug_select;
 	    shutdown_thing($c);
 	    next;
 	}
@@ -766,10 +782,10 @@ while (1) {
 		} elsif (ref($new) eq 'IO::Socket::UNIX') {
 		    $peer = "some unix socket";#$new->peerpath();
 		} else {
-		    print "new client, but ref is ",ref($new),"\n";
+		    print STDERR "new client, but ref is ",ref($new),"\n";
 		    $peer = "unknown";
 		}
-		print "new client from $peer on lsocket $l->[0]\n";
+		print STDERR "new client from $peer on lsocket $l->[0]\n";
 	    }
 	    my $new_client = { 'fd' => $new, 'type' => 'client',
 			       'canwrite' => $l->[1]};
@@ -790,7 +806,7 @@ while (1) {
 		# examine a command from a client
 		do_command($c);
 	    } else {
-		print "read ",length($buf)," bytes from gatherer ",
+		print STDERR "read ",length($buf)," bytes from gatherer ",
 		  $c->{'source'},"\n" if $debug_gather;
 	    }
 	}
@@ -801,7 +817,7 @@ while (1) {
     my(@old) = grep { $_->{'expire'} < $now } @gatherers;
     foreach (@old) {
 	# expire the sucker
-	print "expiring source ",$_->{'source'},"\n" if $debug_gather;
+	print STDERR "expiring source ",$_->{'source'},"\n" if $debug_gather;
 	shutdown_thing($_);
     }
 }
@@ -812,8 +828,9 @@ sub do_command {
     while ($c->{'incoming'}) {
 	# check to see if there is a valid command in the input buffer
 	my(@cmd) = unpack("C*", $c->{'incoming'});
-	printf("do_command $c: 0x%02x\n",$cmd[0]) if $debug_client;
-	print "total_bytes_sucked: $total_bytes_sucked\n" if $debug_client;
+	print STDERR ("do_command $c: 0x%02x\n",$cmd[0]) if $debug_client;
+	print STDERR "total_bytes_sucked: $total_bytes_sucked\n"
+	  if $debug_client;
 	if ($cmd[0] == 0x00) {
 	    # 1 byte
 	    substr($c->{'incoming'}, 0, 1) = "";
@@ -824,7 +841,7 @@ sub do_command {
 	    if ($cmd[1] * 8 > $pool_entropy) {
 		$count = int($pool_entropy / 8);
 	    }
-	    print " wanted $cmd[1], given $count\n" if $debug_client;
+	    print STDERR " wanted $cmd[1], given $count\n" if $debug_client;
 	    $c->{'outgoing'} .= pack("C",$count) . get_entropy($count);
 	    substr($c->{'incoming'}, 0, 2) = "";
 	} elsif ($cmd[0] == 0x02) {
@@ -837,11 +854,12 @@ sub do_command {
 	    $c->{'outgoing'} .= get_entropy($count);
 	    substr($c->{'incoming'}, 0, 2) = "";
 	    $c->{'wanted'} -= $count;
-	    print "read blocked, they wait for $c->{'wanted'} bytes\n"
+	    print STDERR "read blocked, they wait for $c->{'wanted'} bytes\n"
 	      if $debug_client and $c->{'wanted'};
 	    push(@want_entropy, $c) if $c->{'wanted'};
 	} elsif ($cmd[0] == 0x03 and $c->{'canwrite'}) {
-	    print "cmd: ",unpack("H*",$c->{'incoming'}),"\n" if $debug_client;
+	    print STDERR "cmd: ",unpack("H*",$c->{'incoming'}),"\n"
+	      if $debug_client;
 	    last unless length($c->{'incoming'}) >= 4;
 	    my($dummy,$bits,$bytes) = unpack("CnC", 
 					     substr($c->{'incoming'},0,4));
@@ -853,7 +871,7 @@ sub do_command {
 	    $c->{'outgoing'} .= pack("C",length($pidstring)) . $pidstring;
 	    substr($c->{'incoming'}, 0, 1) = "";
 	} else {
-	    print "bogus client request $cmd[0]\n" if $debug;
+	    print STDERR "bogus client request $cmd[0]\n" if $debug;
 	    $c->{'incoming'} = ""; # just in case
 	    shutdown_thing($c);
 	}
